/* File: VMS_TDB_ISAM.C
 *
 *
 * ??-???-????	Jean-Yves Collot?	SAMBA 2.2.8
 * 19-Sep-2004	J. Malmberg		Update for VMS 8.2
 *==========================================================================*/
/*
 * Special VMS version of TDB persistent storage functions.  This version
 * uses an indexed file with the hash value as its key.
 *
 * The data records in the indexed files have the following layout:
 *
 *    |key_hash|key_len|data_len|key...|data...|
 *
 * where:
 *    key_hash		32-bit integer.  Contains a hash of the key field, the
 *			hash function used varies with the hash version:
 *			    0 - use original gdbm hash function with no mask.
 *			    1 - use VMS LNM hash, masking to 'n' bits, where
 *				'n' is stored in the file's key name field.
 *
 *    key_len		32-bit integer.  Length of key.  Total length of key
 *			and data cannot exceed 2036 bytes.
 *
 *    data_len		32-bit integer.  Length of data bound to key.  Total
 *			length of key and data cannot exceed 2036 bytes.  A
 *			record not long enough to contain the data_len field
 *			(12 bytes) is considered a 'deleted' record.
 *
 *    key		Tag used by application for storing an retrieving data
 *			from the tdb.
 *
 *    data		Data associated with tag.
 *
 * The key name in the prolog 3 key descriptor is used to determine the key
 * version.  If the name is blank or missing, version 0 is assumed, which
 * means the file was created with a previous version of VMS_TDB.C.  Otherwise,
 * the field is of the form "TDBHASH.v/arg" where 'v' is the version number
 * (e.g. 1) and 'arg' is an argument specific to the version (e.g. mask size).
 * The default mask size for a new file is 9 bits (512 keys), but this may
 * be overridden by an entry of the file in vms_tdb_default.conf if present in
 * the same directory.
 *
 * When the key version is greater than 0, deleted records are not deleted but
 * simply truncated down to the hash_key field.  Insert operations will
 * reuse a deleted record with the same hash key if possible.  The reuse limits
 * the impact application access patterns that create and delete lots of
 * entries, which would otherwise fill the data buckets with deleted records.
 * With reuse, the number of null entries is more or less one per allowed
 * key value (determined by mask).
 *
 * The RMS calls specify manual unlocking of the records, so all records
 * in a hash chain will become locked prior to release.
 *
 * Date:	29-OCT-2003
 */
#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <starlet.h>
#include <fscndef.h>
#include <ssdef.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <unixio.h>		/* getname */
#include <unixlib.h>
#include <fdldef.h>			/* File description language */
#include <lib$routines.h>		/* lib$ routines (find_image_symbol) */
#include <rms.h>
/* #include <rmsdef.h> */
#include <lckdef.h>
#include <descrip.h>
/* #include <prvdef.h> */
/* #include <lkidef.h> */
/* #include <signal.h> */
/* typedef volatile sig_atomic_t SIG_ATOMIC_T; */

#pragma message disable dollarid

#include "tdb.h"
/* #include "spinlock.h" */

static 	u32 null_hash = 0;
static u32 hash;

/* NB assumes there is a local variable called "tdb" that is the
 * current context, also takes doubly-parenthesized print-style
 * argument. */
#define TDB_LOG(x) (tdb->log_fn?((tdb->log_fn x),0) : 0)

#define CONFIG_FILENAME "vms_tdb_defaults.conf"
#define DEFAULT_MASK_SIZE 9

extern TDB_DATA tdb_null;

int smb_panic(const char * string);
#ifdef DEBUG_DISP
#define TDB_DEBUG 1
static void disp_key ( TDB_DATA key, const char *prefix, TDB_CONTEXT *tdb )
{
    int i, j, length;
    TDB_ISAM *isam = tdb->isam;
    char disp[2048];

    length = key.dsize < 180 ? key.dsize : 179;
    if ( key.dptr ) {
	for ( i = j = 0; i < length; i++ ) {
	    sprintf ( &disp[j], "%02x", key.dptr[i] ); j+= 2;
	    if ( (i&3) == 3 ) if ( (i+1) < length) disp[j++] = '-';
        }
	disp[j] = '\0';
    } else {
	strcpy ( disp, "<NULL>" );
    }
    printf ( "%s key: size=%d, data= '%s'\n", prefix, key.dsize, disp );
    printf ( "   tdb address: %x (%s), rab id: %d\n", tdb,
	tdb->name ? tdb->name : "", isam->rab.rab$b_bid );
    if ( !isam->rab.rab$b_bid )
	smb_panic ( "Invalid RAB block in non-internal TDB" );
}
#else
#define disp_key(x,y,z) if ( !z->isam->rab.rab$b_bid && !z->hash_mask ) \
  smb_panic ( "Invalid RAB block in non-internal TDB\n" );
#endif

/*
 * Hash key data based on algorithm specified by version argument.
 */
static u32 tdb_hash(TDB_DATA key, int version, u32 mask)
{
    u32 value;	/* Used to compute the hash value.  */
    u32   i;	/* Used to cycle through random values. */
    u32 octet;

    if ( (key.dsize == 0) || (!key.dptr) ) return 0;

    switch ( version ) {
	case 0:
	    /* Use gdbm hash algorithm for compatibility with previous TDB. */
	    /* Set the initial value from the key size. */
	    for (value = 0x238F13AF * key.dsize, i=0; i < key.dsize; i++)
		value = (value + (key.dptr[i] << (i*5 % 24)));

	    value = (1103515243 * value + 12345);
	    break;

	case 1:
	    /*
	     * Use good 'ole VMS logical name table hash algorithm.
	     */
	    value = 0;
	    for ( i = 0; i < key.dsize; i++ ) {
	    octet = (u32) key.dptr[i];
		value = mask & (octet*577 ^ value*3);
	    }
	    break;

	default:
	    fprintf(stderr,
		"BUGCHECK, tdb file has unsupported key version: %d\n", version);
	    value = 0;
	    break;
    }
    /* printf ( "hash key of length %d, result: %x\n", key.dsize, value ); */
    return value;
}
/*
 * The global access lock functions support the clear_if_first option.
 */
static struct smblock {
    struct smblock *next;
    int sts;
    int val;
    char resnam[40];
} *smblocks;

static int global_access_lock (TDB_CONTEXT *tdb, int rw_type, int *result )
{
    int status, length, flags, mode;
    char resnam[40], *tmp;
    struct smblock *lock, *prev_lock, provisional_lock;
    struct { int l; char *s; } resnam_dx;
    /*
     * Construct lock resource name, max allowed is 31 characters.
     * We really should be using a NAM block with the open and contructing
     * a unique name from the device and file id.
     */
    tmp = strrchr (tdb->name,'/');
    if ( !tmp ) tmp = tdb->name;		/* no '/' in name */
    else tmp++;					/* skip the / */
    length = strlen ( tmp );
    if ( length > 24 ) length = 24;
    strncpy ( resnam, "SMBTDB_", 7 );
    strncpy ( &resnam[7], tmp, length );
    resnam[length+7] = '\0';
    /*
     * Search lock list to see if a lock with this resource name already
     * acquired.  Keep track of previous entry in list for easy deletion.
     */
    prev_lock = 0;
    for ( lock = smblocks; lock; lock = lock->next ) {
	if ( strcasecmp(resnam,lock->resnam) == 0 ) break;
	prev_lock = lock;
    }
    /*
     * Set up flags for ENQ operation.
     */
    if ( lock ) {
	flags = LCK$M_CONVERT;		/* lock exists */
    } else {
	/*
	 * Lock not in list, use a temporary structure.
	 */
	flags = 0;
	lock = &provisional_lock;	/* automatic storage */

	memcpy ( lock->resnam, resnam, 32 );
	lock->next = 0;

    }
    /*
     * Take initial action based on lock operation requested.
     */
    switch ( rw_type ) {
    case F_UNLCK:
	if ( lock ) {
	    /*
	     * Lock present, dequeue and remove from lock list.
	     */
	    status = sys$deq ( lock->val, 0, 0, 0 );
	    if ( prev_lock ) prev_lock->next = lock->next;
	    else smblocks = lock->next;		/* was first in list */

	    free ( lock );
	    lock = 0;
	}
	if ( result ) *result = F_UNLCK;
	break;

    case F_RDLCK:
	/*
 	 * Get a PR (protected read) mode lock or convert existing one.
	 */
 	resnam_dx.l = strlen ( lock->resnam );
	resnam_dx.s = lock->resnam;

	status = sys$enqw ( 0, LCK$K_PRMODE, &lock->sts,
                     LCK$M_SYSTEM | flags,
                     &resnam_dx,
                     0, 0, 0, 0, 0, 0, 0, 0);
	if ( (status&1) == 1 ) status = lock->sts;
	if ( result ) *result = F_RDLCK;

	break;

    case F_WRLCK:
	/*
 	 * Get a CR (concurrent read) mode lock or convert existing one.
	 */
 	resnam_dx.l = strlen ( lock->resnam );
	resnam_dx.s = lock->resnam;

	status = sys$enqw ( 0, LCK$K_CRMODE, &lock->sts,
                     LCK$M_SYSTEM | flags,
                     &resnam_dx,
                     0, 0, 0, 0, 0, 0, 0, 0);
	if ( (status&1) == 1 ) status = lock->sts;

	break;

    default:
	/*
	 * Bad rw_type.
	 */
	status = SS$_BADPARAM;
	break;
    }
    /*
     * Set error number and abort if lock operation failed.
     */
    if ( (status&1) == 0 ) {
	errno = EVMSERR;
	vaxc$errno = status;
	if ( result ) *result = F_UNLCK;
	return 0;
    }
    /*
     * Make a provisional lock permanent, copy to new structure and add to list.
     */
    if ( lock == &provisional_lock ) {
	lock = malloc ( sizeof(struct smblock) );
	if ( !lock ) {
	    return 0;
	}
	*lock = provisional_lock;

	lock->next = smblocks;
	smblocks = lock;
    }
    /*
     * Take additional action if rw_type is W_WRLCK.  Attempt to convert
     * the CR mode lock to PW (protected write) mode with the NOQUEUE option.
     * Other processes with a CR mode lock will not block the conversion but
     * any holding a PR mode lock will cause the request to fail (leaving us
     * with a CR mode lock).
     */
    if ( rw_type == F_WRLCK ) {

	if ( result ) *result = F_WRLCK;

	status = sys$enqw ( 0, LCK$K_EXMODE, &lock->sts,
	    LCK$M_SYSTEM | LCK$M_CONVERT | LCK$M_NOQUEUE,
	    0, 0, 0, 0, 0, 0, 0, 0, 0 );

	if ( status == SS$_NOTQUEUED ) {
	    /*
	     * Another guy got a PW lock on the resource before we did. Upgrade
	     * our CR lock to PR mode, which forces us to wait for the other
	     * guy to complete initialization.
	     */
	    return global_access_lock ( tdb, F_RDLCK, result );

	} else if ( (status&1) == 1 ) {
	    /*
	     * Lock didn't block, verify that we got the lock.  If we did,
	     * leave result at F_WRLCK to tell caller we have the PW mode lock.
	     */
	    if ( (lock->sts & 1) == 0 ) {
		errno = EVMSERR;
		vaxc$errno = lock->sts;
	 	if ( result ) *result = F_UNLCK;
	       return 0;
	    }
	} else {
	    /*
	     * Error queuing lock.
	     */
	    errno = EVMSERR;
	    vaxc$errno = status;
	    if ( result ) *result = F_RDLCK;
	    return 0;
	}
    }
    return 1;
}

/*
 * The FDL for creating the index file.
 */
static $DESCRIPTOR(tdb_fdl_dx, "");
static char *tdb_base_fdl = "\
  FILE;	GLOBAL_BUFFER_COUNT 6; ORGANIZATION indexed;\
    PROTECTION (system:RWD, owner:RWD, group:RW, world:RW);\
  RECORD; CARRIAGE_CONTROL carriage_return; FORMAT variable; SIZE 0;\
  AREA 0; ALLOCATION 999; BEST_TRY_CONTIGUOUS yes; BUCKET_SIZE 12;\
    EXTENSION 1296;\
  AREA 1; ALLOCATION 24; BEST_TRY_CONTIGUOUS yes; BUCKET_SIZE 12;\
    EXTENSION 12;\
  KEY 0; CHANGES no; DATA_AREA 0; DATA_FILL 100; DATA_KEY_COMPRESSION no;\
    DATA_RECORD_COMPRESSION yes; DUPLICATES yes; INDEX_AREA 1;\
    INDEX_COMPRESSION no; INDEX_FILL 100; LEVEL1_INDEX_AREA 1; PROLOG 3;\
    SEG0_LENGTH 4; SEG0_POSITION 0; TYPE bin4;";
/*
 * FDL related functions.  Encode the mask size (in number of bits) in the key
 * name assigned to the file and on subsequent opens decode this size and set
 * key_mask field appropiately.  We store the information in the key name field
 * of the primary key, using a name of the form "TDBHASH.v/arg", where 'v'
 * is the hash algorithm version (1) and arg is an algorithm-dependant
 * parameter for the algorithm.
 *
 * The 32-byte key name field is not used by VMS and intended to provide
 * descriptive text about the purpose of the index.
 */
static struct dsc$descriptor_s *fdl_string ( int mask_size )
{
    int length;
    length = strlen ( tdb_base_fdl );
    tdb_fdl_dx.dsc$a_pointer = malloc ( length + 64 );
    sprintf ( tdb_fdl_dx.dsc$a_pointer, "%s NAME \"TDBHASH.1/%d\";",
	tdb_base_fdl, mask_size );
    tdb_fdl_dx.dsc$w_length = strlen ( tdb_fdl_dx.dsc$a_pointer );
    return &tdb_fdl_dx;
}
/*
 * Parse key name file in XABKEY block loaded when we opened the file and set
 * tdb->isam->key_version and tdb->key_mask accordinging.  If the key field is
 * not in the proper form, leave the fields untouched (default value is
 * algorithm 0).
 */
static int default_mask_size_hint;
static void set_key_mask ( TDB_ISAM *isam )
{
    char name[33],term, *slash;
    int i, size, version;
    u32 mask;
    char *keyptr = (char*)isam->key.xab$l_knm;

    /*
     * Copy name field up to first space or null.  VMS blank-pads the field.
     */
    for ( i = 0; (i < 32) && *keyptr && (*keyptr != ' '); i++)
    {
	name[i] = *keyptr++;
    }
    name[i] = '\0';
    /*
     * Parse name field to extract version and argument (mask size).
     */
    if ( strncmp ( name, "TDBHASH.", 8 ) != 0 ) return;
    slash = strchr ( &name[8], '/' );
    if ( !slash ) return;		/* slash not found */
     *slash = '\0';
    version = atoi ( &name[8] );
    if ( (version < 0) || (version > 1) ) {
	fprintf(stderr,"BUGCHECK, tdb file hash unsupported version\n" );
	return;
    }
    size = atoi ( &slash[1] );
    if ( size < 0 ) return;
    /*
     * Generate mask.  force 0 bit and 32 bit size to algorithm 0.
     */
    if ( (size < 1) || (size > 31) ) version = 0;
    isam->key_version = version;
    for ( mask = i = 0; i < size; i++ ) mask = (mask<<1) | 1;
    isam->key_mask = (version == 0) ? 0xffffffff : mask;
#ifdef TDB_DEBUG
    printf ( "Set hash mask for primary key to %x\n", isam->key_mask );
#endif
}
/*
 * Determine default mask size to use for a given file name by searching
 * for that name in a configuration file stored in the same directory.
 */
static int scan_config_file ( FILE *cf, char *fname, int fn_len )
{
    char line[4096], *nl, *tok[20];
    int size, length, tcnt;

    size = default_mask_size_hint;
    length = 0;
    while ( fgets ( &line[length], sizeof(line)-length, cf ) ) {
	/*
	 * Trim newline and process continuation lines.
	 */
	nl = strchr ( &line[length], '\n' );
	if ( nl ) *nl = '\0';
	else nl = &line[strlen(line)];
	length = nl - line;
	if ( (length > 0) && (length < sizeof(line)) &&
		 (line[length-1] == '\\') ) {
	    line[length-1] = ' ';
	    continue;
	}
	length = 0;
	/*
	 * Parse first element from line and continue with next line if
	 * it is a comment line or doesn't start with the target file name.
	 * Otherwise, build array of remaining line elements.
	 */
	tok[0] = strtok ( line, " \t" );
	if ( !tok[0] ) continue;
	if ( tok[0][0] == '!' ) continue;	/* comment line */
	if ( tok[0][0] == ';' ) continue;
	if ( tok[0][0] == '#' ) continue;
	if ( (tok[0][0] != '*') &&
		strncasecmp ( tok[0], fname, fn_len ) != 0 ) continue;

	for ( tcnt = 0; tok[tcnt]; tcnt++ ) {
	    if ( tcnt >= 19 ) break;
	    tok[tcnt+1] = strtok ( 0, " \t" );
	}
	/*
	 * Process the command line elements.
	 */
	if ( tcnt > 1 ) {
	    size = atoi ( tok[1] );
	}
    }
    return size;
}


static int default_mask_size ( struct dsc$descriptor_s *fname_dx )
{
    int status, flags, length, size;
    struct { short length, code; char *s; } vlist[5];
    char *name, conf_fname[512];
    FILE *cf;
    /*
     * Parse the VMS filename to find starting address of name portion.
     */
    vlist[0].code = FSCN$_NAME;
    vlist[1].code = FSCN$_TYPE;
    vlist[2].code = vlist[2].length = 0;

    status = sys$filescan ( fname_dx, vlist, &flags, 0, 0 );
    if ( (status&1) == 0 ) {
	return default_mask_size_hint;	/* filename not valid, use default size */
    }
    name = vlist[0].s;
    if ( !name ) name = vlist[1].s;
    if ( !name ) name = &fname_dx->dsc$a_pointer[fname_dx->dsc$w_length];
    /*
     * Construct filename for configuration file.
     */
    length = (name - fname_dx->dsc$a_pointer);
    if ( length > (sizeof(conf_fname)-30) ) return default_mask_size_hint;
    strncpy ( conf_fname, fname_dx->dsc$a_pointer, length );
    strcpy ( &conf_fname[length], CONFIG_FILENAME );
    cf = fopen ( conf_fname, "r" );
    size = default_mask_size_hint;
    if ( cf ) {
	/*
	 * Read contents for file.
	 */
	length = vlist[0].length + vlist[1].length;
	size = scan_config_file ( cf, vlist[0].s, length );
	fclose ( cf );
    }
    return size;			/* default value, 512 keys. */
}
/*
 * Action routines for decc$to_vms that verify the existence of target
 * file, saving its VMS-style name in global variable to_vms_filename[];
 */
static char to_vms_filename[1024];
static int verify_existing ( char *vms_name, int is_directory )
{
    /*
     * See if file exists and is a keyed file.  if so, set to_vms_filename.
     * Note that stat call sets errno on failure for reporting purposes.
     */
    struct stat info;
    int rfm, length;

    to_vms_filename[0] = '\0';

    if ( stat ( vms_name, &info )== 0 ) {
	/*
	 * Should test for file organization, for now use record format.
	 * Files that we didn't create are most likely stream files from
	 * the standard memory-based TDB functions.
	 */
	rfm = info.st_fab_rfm;
	if ( rfm != FAB$C_VAR ) return 0;
	/*
	 * File is good, save its name for open later with SYS$OPEN().
	 */
	length = strlen ( vms_name );
	if ( length >= sizeof ( to_vms_filename ) ) return 0;
	strcpy ( to_vms_filename, vms_name );

	return 1;
    }
    return 0;
}

static int verify_existing_or_create ( char *vms_name, int is_directory )
{
    int status;
    unsigned long flags;
    int length, hash_mask_size;
    static $DESCRIPTOR(filename,"");
    static $DESCRIPTOR(fdlshr,"FDLSHR");
    static $DESCRIPTOR(fdlcreate,"FDL$CREATE");
    static int images_loaded = 0;
    static int (*fdl_create)
       (const struct dsc$descriptor_s *fdl_string,
	const struct dsc$descriptor_s *filename,
	const struct dsc$descriptor_s *default_name,
	struct dsc$descriptor_s *result_name,
	unsigned long * fdl_block,
	const unsigned long * flags,
	unsigned long * stmnt_num,
	unsigned long * retlen,
	unsigned long * sts,
	unsigned long * stv,
	const struct dsc$descriptor_s *default_fdl);
    /*
     * First to verify part, we are done if file exists.  Otherwise
     * delete existing bad-format file.
     */
    status = verify_existing ( vms_name, is_directory );
    if ( status != 0 ) return status;
    delete ( vms_name );
    /*
     * Load FDL code dynamically since we expect to rarely need it.
     */
    if ( !images_loaded ) {
	status = lib$find_image_symbol ( &fdlshr, &fdlcreate, &fdl_create );
    	if ( (status&1) == 0 ) {
	    errno = EVMSERR;
	    vaxc$errno = status;
	    return 0;
	}
	images_loaded = 1;
    }
    /*
     * Set descriptors to point to passed strings and call FDL$CREATE.
     */
    filename.dsc$w_length = strlen ( vms_name );
    filename.dsc$a_pointer = vms_name;
    hash_mask_size = default_mask_size ( &filename );	/* default value */

    flags = FDL$M_FDL_STRING | FDL$M_SIGNAL;

    status = fdl_create
	       (fdl_string(hash_mask_size),
		&filename,
		0,	/* default name */
		0,	/* result name */
		0,	/* fdl_block */
		&flags, /* flags */
		0,	/* stmnt_num */
		0,	/* retlen */
		0,	/* sts */
		0,	/* stv */
		0);	/* default fdl */

    /*
     * Synthesize C errno behaviour on error, save creatd name on success.
     */
    if ( (status&1) == 0 ) {
	errno = EVMSERR;
	vaxc$errno = status;
    } else {
	length = strlen ( vms_name );
	if ( length >= sizeof ( to_vms_filename ) ) return 0;
	strcpy ( to_vms_filename, vms_name );
    }
    return (status&1);
}

static TDB_DATA set_tdb_data ( TDB_CONTEXT *tdb, TDB_ISAM *isam )
{
    isam->rec_valid = TRUE;
    isam->last_hash = isam->tdb_record.hash_key;
    tdb->tdb_dkey.dsize = isam->tdb_record.key_size;
    tdb->tdb_dkey.dptr = isam->tdb_record.value;
    tdb->tdb_dbuf.dsize = isam->tdb_record.data_size;
    tdb->tdb_dbuf.dptr = &isam->tdb_record.value[tdb->tdb_dkey.dsize];
#ifdef TDB_DEBUG
    printf ( "   setting dbuf: l=%d p=%x\n", tdb->tdb_dbuf.dsize,
		tdb->tdb_dbuf.dptr );
#endif
    return tdb->tdb_dbuf;
}
/*
 * Internal function for lookup and load of a record from the file.  All
 * records read from hash chain remain locked if keep_lock argument is true,
 * use $FREE service to release them.  Keeping all records locked avoids
 * some race conditions.  Note that a failed load may still leave one or
 * more records locked (avoids race in adding record).
 */
static TDB_DATA load_record (TDB_CONTEXT *tdb, TDB_DATA key, int keep_lock)
{
    char found = FALSE;
    int sts, *kv, count;
    u32 hash;
    TDB_ISAM *isam = tdb->isam;
    /*
     * Return previous record if it still matches (should check locking).
     */
    hash = tdb_hash ( key, isam->key_version, isam->key_mask );
#ifdef TDB_DEBUG
    printf ( "load_rec hash: %x, last_hash: %x valid: %d\n",
	hash, isam->last_hash, isam->rec_valid );
#endif
    if (isam->rec_valid && (isam->last_hash == hash) )
    {
	if ( (key.dsize == tdb->tdb_dkey.dsize) && key.dptr &&
	    (memcmp ( key.dptr, tdb->tdb_dkey.dptr, key.dsize ) == 0) ) {
	    return tdb->tdb_dbuf;
	}
    }
    /*
     * Read keyed entry in file.
     */
    isam->rec_valid = FALSE;			/* handing buffer to RMS */
    isam->last_hash = hash;
    isam->last_deleted_valid = 0;
    isam->rab.rab$l_kbf = (char *) &hash;
    isam->rab.rab$b_krf = 0;
    isam->rab.rab$b_ksz = sizeof(hash);
    isam->rab.rab$l_ubf = (char *) &isam->tdb_record;
    isam->rab.rab$w_usz = sizeof(isam->tdb_record);
    isam->rab.rab$w_rsz = 0;
    if ( keep_lock ) {
	isam->rab.rab$v_kge = 1;
    } else {
	isam->rab.rab$v_kge = 0;		/* exact match only */
    }
    isam->rab.rab$b_rac = RAB$C_KEY;
    sts = sys$get(&isam->rab);
#ifdef TDB_DEBUG
    printf ( "   status of initial get: %d (l=%d,k=%x), rab %x hdr; %x %d\n", sts,
      isam->rab.rab$w_rsz, isam->tdb_record.hash_key, &isam->rab, isam->rab.rab$b_bid, isam->rab.rab$b_bln );
#endif
    /*
     * Compare key field in successive records until a match.
     */
    count = 0;				/* number of locked records */
    while ( (sts&1) == 1 ) {
	count++;
	if ( (isam->tdb_record.hash_key != hash) &&
			(isam->rab.rab$w_rsz >= 4) ) {
	    /*
	     * Record read, but hash key different than target.
	     */
	    break;
	} else if ( isam->rab.rab$w_rsz <= 12 ) {
	    /*
	     * Record read, but to short to have length fields, indicated it
	     * is a 'deleted record.  Save RFA of record so we can reuse.
	     */
	    isam->last_deleted_valid = 1;
	    isam->last_deleted_hash = hash;
	    memcpy ( isam->last_deleted_rfa, &isam->rab.rab$l_rfa0,
			sizeof(isam->last_deleted_rfa) );

        } else if (isam->tdb_record.key_size == key.dsize) {
	    /*
	     * Key value right length, see if strings match.
	     */
            if (memcmp(isam->tdb_record.value, key.dptr, key.dsize) == 0)
            {
                    found = TRUE;
		    break;
            }
        }
	/*
	 * Read next record sequentially.
         */
        isam->rab.rab$b_rac = RAB$C_SEQ;	    /* Read next record */
        sts = sys$get(&isam->rab);
#ifdef TDB_DEBUG
	printf ( "   status of secondary get: %d\n", sts );
#endif
    }
    /*
     * Cleanup RMS state and save result if found.
     */
    if ( (count > 0) && (keep_lock==0) ) sys$free ( &isam->rab );

    if (found) return set_tdb_data ( tdb, isam );

    return tdb_null;
}


/* find an entry in the database given a key */
static TDB_DATA isam_fetch(TDB_CONTEXT *tdb, TDB_DATA key)
{
    int sts;

    disp_key ( key, "Fetchin, ", tdb );
    tdb->tdb_dbuf = load_record (tdb, key, 0);
    return tdb_dupe(tdb->tdb_dbuf);
}

/* traverse the entire database - calling fn(tdb, key, data) on each element.
   return -1 on error or the record count traversed
   if fn is NULL then it is not called
   a non-zero return value from fn() indicates that the traversal should stop
  */
static int isam_traverse(TDB_CONTEXT *tdb, tdb_traverse_func fn, void *state)
{
	TDB_DATA key, dbuf;
	int count = 0;
	int sts;
	TDB_ISAM *isam = tdb->isam;

	isam->rab.rab$b_rac = RAB$C_KEY;
	isam->rab.rab$v_kge = 1;
	isam->rab.rab$l_kbf = (char *) &null_hash;
	isam->rab.rab$w_usz = sizeof(isam->tdb_record);
	isam->rab.rab$w_rsz = 1000;
	isam->rec_valid = FALSE;

	for ( sts = sys$get(&isam->rab);	 (sts&1) == 1;
		sts = sys$get(&isam->rab) ) {

	    if ( isam->rab.rab$w_rsz >= 12 ) {
	      count++;
	      key.dsize = isam->tdb_record.key_size;
	      key.dptr = &isam->tdb_record.value[0];
	      dbuf.dsize = isam->tdb_record.data_size;
	      dbuf.dptr = &isam->tdb_record.value[key.dsize];
	      if ( isam->tdb_record.hash_key > isam->key_mask ) {
		fprintf(stderr,
		   "BUGCHECK, tdb file %s has RMS key of %x when mask is %x\n",
		    tdb->name, isam->tdb_record.hash_key, isam->key_mask );
		disp_key(key, "record", tdb);
		    smb_panic ( "fatal bugcheck" );
	      }
	      sys$release ( &isam->rab );
	      disp_key ( key, "traverse visiting", tdb );
	      if (fn && fn(tdb, key, dbuf, state)) {
		/* They want us to terminate traversal */
		sys$free (&isam->rab);
		return count;
	      }
	    } else sys$release ( &isam->rab );
	    isam->rab.rab$w_usz = sizeof(isam->tdb_record);
	    isam->rab.rab$w_rsz = 1000;
	    isam->rab.rab$b_rac = RAB$C_SEQ;
	}
	/* printf ( "Final get status: %d, count: %d\n", sts, count ); */

	return count;
}

/* find the first entry in the database and return its key */
static TDB_DATA isam_firstkey(TDB_CONTEXT *tdb)
{
    int sts;
    TDB_ISAM *isam = tdb->isam;

    isam->rab.rab$b_rac = RAB$C_KEY;
    isam->rab.rab$v_kge = 1;
    isam->rab.rab$l_kbf = (char *) &null_hash;
    isam->rab.rab$b_ksz = sizeof(null_hash);
    isam->rab.rab$l_ubf = (char *) &isam->tdb_record;
    isam->rab.rab$w_usz = sizeof(isam->tdb_record);
    isam->rec_valid = FALSE;
    for ( sts=sys$get(&isam->rab); (sts&1) == 1; sts = sys$get(&isam->rab) ) {
	if ( isam->rab.rab$w_rsz >= 12 ) {
	    sys$free(&isam->rab);
	    set_tdb_data ( tdb, isam );
	    return tdb_dupe ( tdb->tdb_dkey );
	}
	isam->rab.rab$b_rac = RAB$C_SEQ;
    }
    sys$free ( &isam->rab );
    return tdb_null;
}

/* find the next entry in the database, returning its key */
static TDB_DATA isam_nextkey(TDB_CONTEXT *tdb, TDB_DATA oldkey)
{
    int sts;
    TDB_ISAM *isam = tdb->isam;

    isam->rab.rab$b_rac = RAB$C_SEQ;
    isam->rab.rab$l_ubf = (char *) &isam->tdb_record;
    isam->rab.rab$w_usz = sizeof(isam->tdb_record);
    isam->rec_valid = FALSE;
    for ( sts=sys$get(&isam->rab); (sts&1) == 1; sts = sys$get(&isam->rab) ) {
	if ( isam->rab.rab$w_rsz >= 12 ) {
	    sys$free(&isam->rab);
	    set_tdb_data ( tdb, isam );
	    return tdb_dupe ( tdb->tdb_dkey );
	}
    }
    sys$free(&isam->rab);
    return tdb_null;
}

/* delete an entry in the database given a key */
static int isam_delete(TDB_CONTEXT *tdb, TDB_DATA key)
{
    TDB_DATA data;
    int sts;
    TDB_ISAM *isam = tdb->isam;

    disp_key ( key, "Deleting, ", tdb );

    isam->rec_valid = 0;		/* force read */
    data = load_record (tdb, key, 1);
    if (data.dptr != NULL)
    {
	/*
	 * For comatibility, only use record truncation technique if key
	 * version higher than 0.  Truncate record to only contain the hash.
	 */
	isam->rec_valid = FALSE;
	if ( isam->key_version > 0 ) {
	    /* Perform delete by truncating record */
	    isam->rab.rab$w_rsz = 4;
	    sts = sys$update ( &isam->rab );
	} else {
	    sts = sys$delete(&isam->rab);
	}
	if ((sts & 1) != 1)
                TDB_LOG((tdb, 0, "$delete failed (%s)\n", strerror(EVMSERR,sts)));
    }
    sys$free ( &isam->rab );
    return 0;
}

/* store an element in the database, replacing any existing element
   with the same key

   return 0 on success, -1 on failure
*/
static int isam_store(TDB_CONTEXT *tdb, TDB_DATA key, TDB_DATA dbuf, int flag)
{
    int sts, sd_dsize;
    int tobeupdated = FALSE;
    int found_deleted;
    TDB_ISAM *isam = tdb->isam;
    char save_data[sizeof(isam->tdb_record.value)];

    disp_key ( key, "Storing, ", tdb );
    /*
     * Copy user's data buffer to local storage, don't know why (avoid
     * overlap issues with tdb_record.value?).
     */
    sd_dsize = dbuf.dsize;	/* size of saved data */
    if ( (dbuf.dsize + key.dsize) > sizeof(isam->tdb_record.value) ) {
	fprintf(stderr,"Record for file '%s' is too big (%d+%d)\n",
		isam->fab.fab$l_fna, key.dsize, dbuf.dsize );
	sd_dsize = sizeof(isam->tdb_record.value) - key.dsize;
    }
    memcpy(save_data,dbuf.dptr,sd_dsize);

    isam->rec_valid = 0;
    tdb->tdb_dbuf = load_record (tdb,key,1);

    found_deleted = FALSE;
    if (tdb->tdb_dbuf.dptr != NULL)
	tobeupdated = TRUE;
    else if ( (isam->key_version > 0) && (isam->last_deleted_valid) ) {
	/*
	 * Refetch and lock a 'deleted' record, verify it is still deleted.
	 */
	memcpy ( &isam->rab.rab$l_rfa0, isam->last_deleted_rfa,
		sizeof(isam->last_deleted_rfa) );
	isam->rab.rab$w_usz = sizeof(isam->tdb_record);
	isam->rab.rab$w_rsz = 0;
	isam->rab.rab$b_rac = RAB$C_RFA;		/* get by record file addr. */
	sts = sys$get ( &isam->rab );
	if ( (sts&1) && (isam->rab.rab$w_rsz < 12) ) found_deleted = TRUE;
    }

    isam->tdb_record.hash_key = isam->last_hash;		/* optimize */
    if ( !tobeupdated && (flag == TDB_MODIFY) ) {
	tdb->ecode = TDB_ERR_NOEXIST;
	sys$free (&isam->rab);
	return -1;
    } else if ( tobeupdated && (flag == TDB_INSERT) ) {
	tdb->ecode = TDB_ERR_EXISTS;
	sys$free (&isam->rab);
	return -1;
    }
    if (!tobeupdated && !found_deleted )
    {
	isam->rab.rab$b_rac = RAB$C_KEY;
	isam->tdb_record.key_size = key.dsize;
	isam->tdb_record.data_size = dbuf.dsize;
	memcpy(&isam->tdb_record.value[0], key.dptr, key.dsize);
	memcpy(&isam->tdb_record.value[key.dsize],dbuf.dptr,dbuf.dsize);
	isam->rab.rab$w_usz = key.dsize + dbuf.dsize + 12;
	isam->rab.rab$w_rsz = key.dsize + dbuf.dsize + 12;
	sts = sys$put(&isam->rab);
	if ((sts & 1) != 1)
                TDB_LOG((tdb, 0, "$put failed (%s)\n", strerror(EVMSERR,sts)));
    }
    else
    {
	isam->rab.rab$b_rac = RAB$C_KEY;
	isam->tdb_record.key_size = key.dsize;
	isam->tdb_record.data_size = dbuf.dsize;
	memcpy(&isam->tdb_record.value[0], key.dptr, key.dsize);
	memcpy(&isam->tdb_record.value[key.dsize],save_data,sd_dsize);
	isam->rab.rab$w_usz = key.dsize + dbuf.dsize + 12;
	isam->rab.rab$w_rsz = key.dsize + dbuf.dsize + 12;
	sts = sys$update (&isam->rab);
	if ((sts & 1) != 1)
                TDB_LOG((tdb, 0, "$update failed (%s)\n", strerror(EVMSERR,sts)));
    }
    isam->rec_valid = FALSE;
    sys$free ( &isam->rab );
    if ((sts&1) != 1)
    {
	TDB_LOG((tdb, 0, "tdb_store: can't store : %s\n",strerror(EVMSERR,sts)));
	tdb->ecode = TDB_ERR_IO;
	return -1;
    }
    tdb->ecode = TDB_SUCCESS;
    return 0;
}

/* close a database */
static int isam_close(TDB_CONTEXT *tdb)
{
    if ( tdb->isam ) {
	global_access_lock (tdb, F_UNLCK, 0 );
        sys$disconnect(&tdb->isam->rab);
        sys$close(&tdb->isam->fab);
	free ( tdb->isam );
	tdb->isam = (TDB_ISAM *) 0;
    }
    return 0;
}

/* Unlock the keys previously locked by tdb_lockkeys() */
static void isam_unlockkeys(TDB_CONTEXT *tdb)
{
}

/* lock/unlock one hash chain. This is meant to be used to reduce
   contention - it cannot guarantee how many records will be locked */
static int isam_chainlock(TDB_CONTEXT *tdb, TDB_DATA key)
{
   return 0;
}

static int isam_chainunlock(TDB_CONTEXT *tdb, TDB_DATA key)
{
   return 0;
}


/* reopen a tdb - this is used after a fork to ensure that we have an independent
   seek pointer from our parent and to re-establish locks */
static int isam_reopen(TDB_CONTEXT *tdb)
{
    return 0;
}

/* reopen all tdb's */
static int isam_reopen_all(void)
{
    return 0;
}

/* check if an entry in the database exists

   note that 1 is returned if the key is found and 0 is returned if not found
   this doesn't match the conventions in the rest of this module, but is
   compatible with gdbm
*/
static int isam_exists(TDB_CONTEXT *tdb, TDB_DATA key)
{

    tdb->tdb_dbuf = load_record (tdb, key, 0);
    if (tdb->tdb_dbuf.dptr == NULL)
	return 0;
    else
	return 1;
}
static int dump_record ( TDB_CONTEXT *tdb, TDB_DATA key, TDB_DATA value,
	void *state )
{
    printf ( "record hash %x\n", tdb->isam->tdb_record.hash_key );
    return 0;			/* Continue dump */
}
static void isam_dump_all(TDB_CONTEXT *tdb)
{
    isam_traverse ( tdb, dump_record, 0 );
}
static int isam_printfreelist(TDB_CONTEXT *tdb )
{
    printf ( "Not implemented for isam data files\n" );
    return 0;
}


static TDB_METHODS isam_methods = {
    "VMS-indexed",
    isam_fetch, isam_delete, isam_store, isam_close,
    isam_firstkey, isam_nextkey,
    isam_traverse, isam_exists,
    0, 0, 0, 0, 0, 0, isam_dump_all, isam_printfreelist
};
/*
 * Open routine called by tdb_open for non-internal TDBs.  tdb structure
 * has been initialized.
   The open_flags and mode are passed straight to the open call on the
   database file. A flags value of O_WRONLY is invalid. The hash size
   is advisory, use zero for a default value.

   Return is NULL on error, in which case errno is also set.  Don't
   try to call tdb_error or tdb_errname, just do strerror(errno).

 */

int tdb_open_isam ( TDB_CONTEXT *tdb, const char *name, int hash_size,
	mode_t mode )
{
    int count, sts, locked;
    TDB_ISAM *isam;
    /*
     * Load methods.
     */
    tdb->fn = &isam_methods;
    /*
     * Ensure file exists or create a new one.
     */
    if ( hash_size <= 0 ) default_mask_size_hint = DEFAULT_MASK_SIZE;
    else {
	u32 mask;
	for ( default_mask_size_hint = mask = 1; mask < (hash_size-1);
		default_mask_size_hint++ ) mask = (mask<<1) | 1;
    }
    count = decc$to_vms ( name, (tdb->open_flags&O_CREAT) ?
	   	verify_existing_or_create : verify_existing,
	   	0, 0 );		/* don't expand wildcard, disallow directory */

    if ( (count!=1) || (to_vms_filename[0] == '\0') ) {
	TDB_LOG((tdb, 0, "tdb_open: could not translate TDB filename %s: %s\n",
			 name, strerror(EVMSERR,vaxc$errno)));
	return 0;
    }
    /*
     * Initialize RMS structures for opening file.
     */
    isam = malloc ( sizeof(TDB_ISAM) );
    if ( !isam ) return 0;
    tdb->isam = isam;
    isam->fab = cc$rms_fab;
    isam->fab.fab$b_fns = strlen(to_vms_filename);
    isam->fab.fab$l_fna = to_vms_filename;
    isam->fab.fab$b_shr =
		FAB$M_SHRPUT + FAB$M_SHRGET + FAB$M_SHRDEL + FAB$M_SHRUPD;
    isam->fab.fab$l_xab = &isam->key;
    isam->key = cc$rms_xabkey;
    isam->key.xab$l_nxt = 0;			/* end of list. */
    isam->key.xab$b_ref = 0;			/* primary key */
    isam->key.xab$l_knm = isam->key_name;       /* receives key name */

    if ((tdb->open_flags & O_ACCMODE) == O_RDONLY) {
		tdb->read_only = 1;
		/* read only databases don't do locking or clear if first */
		tdb->flags |= TDB_NOLOCK;
		tdb->flags &= ~TDB_CLEAR_IF_FIRST;
		isam->fab.fab$b_fac = FAB$M_GET;
    }
    else
    {
		isam->fab.fab$b_fac = FAB$M_GET + FAB$M_PUT + FAB$M_UPD + FAB$M_DEL;
    }
    /*
     * Open the file and connect record stream.
     */
    sts = sys$open (&isam->fab);
    if ((sts&1) != 1)
    {
	TDB_LOG((tdb, 0, "tdb_open: can't open tdb %s : %s\n",
			 name,
			 strerror(EVMSERR,sts)));
	errno = EINVAL;
	free (isam );
	return 0;
    }
    isam->key_version = 0;
    isam->key_mask = 0xffffffff;
    if ( isam->key.xab$l_knm ) set_key_mask ( isam );
    isam->rec_valid = FALSE;
    isam->last_hash = 0xffffffff;
    isam->last_deleted_valid = 0;

    isam->rab = cc$rms_rab;
    isam->rab.rab$l_rop = RAB$M_WAT + RAB$M_TMO + RAB$M_ULK;
    isam->rab.rab$b_tmo = 8;
    isam->rab.rab$b_rac = RAB$C_KEY;
    isam->rab.rab$l_ubf = (char *) &isam->tdb_record;
    isam->rab.rab$w_usz = sizeof(isam->tdb_record);
    isam->rab.rab$l_rbf = (char *) &isam->tdb_record;
    isam->rab.rab$b_ksz = sizeof(null_hash);
    isam->rab.rab$l_fab = &isam->fab;
    sts = sys$connect (&isam->rab);
    if ((sts&1) != 1)
    {
	TDB_LOG((tdb, 0, "tdb_open: can't connect tdb %s : %s\n",
			 name,
			 strerror(EVMSERR,sts)));
	errno = EINVAL;
	sys$close(&isam->fab);
	free ( isam );
	tdb->isam = 0;
	return 0;
    }

    /* ensure there is only one process initialising at once */
    /*
     * Get a read lock the file or attempt to be the first writer if
         * TDB_CLEAR_IF_FIRST is set.  The locked argument returns
	 * F_WRLCK is we are the first writer or F_RDLCK if not.
	 */
    if ( 0 == global_access_lock ( tdb,
	    (tdb->flags&TDB_CLEAR_IF_FIRST) ? F_WRLCK : F_RDLCK, &locked ) ) {
		TDB_LOG((tdb, 0, "tdb_open: failed to get global lock on %s: %s\n",
			 name, strerror(errno,vaxc$errno)));
		isam_close ( tdb );
		return 0;
    }
    if ( locked == F_WRLCK ) {
	    /*
	     * We are the first writer, clear the database.
	     */
		tdb->open_flags |= O_CREAT;
		isam->rab.rab$b_rac = RAB$C_SEQ;
		isam->rab.rab$w_usz = sizeof(isam->tdb_record);
		isam->rab.rab$w_rsz = 1000;
		sts = 1;
		while ((sts & 1) == 1)
		{
		    sts = sys$get(&isam->rab);
		    if ((sts & 1) == 1)	{
			if ( isam->key_version > 0 ) {
			    isam->rab.rab$w_rsz = sizeof(u32);
			    sys$update ( &isam->rab );
			} else {
			    sys$delete (&isam->rab);
			}
		    }
		}
	    /*
	     * Lower lock to a read lock so others can proceed.
	     */
	    global_access_lock ( tdb, F_RDLCK, 0 );
    }

    return 1;		/* success */
}
